from functools import update_wrapper

from werkzeug.exceptions import Unauthorized

from flask_mysso.context import SecurityContext


def login_required(func):
    def wrapper(*args, **kwargs):
        user = SecurityContext.get_user()
        if user is None:
            raise Unauthorized
        return func(*args, **kwargs)

    return update_wrapper(wrapper, func)


def has_any_role(*roles):
    def decorator(func):
        def wrapper(*args, **kwargs):
            user = SecurityContext.get_user()
            if user is None or not SecurityContext.has_any_role(user, roles):
                raise Unauthorized
            return func(*args, **kwargs)

        return update_wrapper(wrapper, func)

    return decorator


def has_role(role):
    return has_any_role(role)


def has_any_authority(*authorities):
    def decorator(func):
        def wrapper(*args, **kwargs):
            user = SecurityContext.get_user()
            if user is None or not SecurityContext.has_any_authority(user, authorities):
                raise Unauthorized
            return func(*args, **kwargs)

        return update_wrapper(wrapper, func)

    return decorator


def has_authority(authority):
    return has_any_authority(authority)



import json
from typing import Optional

import jwt
import requests

from flask_mysso.context import SecurityContext
from flask_mysso.jwt import JwtManager


class MySSO:
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)
        self._jwt_manager: Optional[JwtManager] = None

    def init_app(self, app):
        app.config.setdefault('MYSSO_CERT', 'http://sso.app.kdsec.org/api/cert')
        resp = requests.get(app.config['MYSSO_CERT'])
        resp.raise_for_status()
        public_key = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(resp.json()))
        self._jwt_manager = JwtManager(public_key=public_key)

        role_prefix = app.config.get('MYSSO_ROLE_PREFIX')
        if role_prefix is not None:
            SecurityContext.DEFAULT_ROLE_PREFIX = role_prefix

        app.before_request(self._handle_token)

    def _handle_token(self):
        self._jwt_manager.authenticate()



__version__ = '0.1'

from .annotation import has_any_authority
from .annotation import has_any_role
from .annotation import has_authority
from .annotation import has_role
from .annotation import login_required
from .context import current_user
from .errors import AuthenticationError
from .errors import InvalidTokenError
from .mysso import MySSO



from json import JSONEncoder as _JSONEncoder

import jwt
from flask import request

from flask_mysso.context import User, SecurityContext
from flask_mysso.errors import InvalidTokenError


class JsonEncoder(_JSONEncoder):
    def default(self, o):
        if isinstance(o, set):
            return list(o)
        return super().default(o)


class JwtHelper:
    @staticmethod
    def decode_jwt(token, key, algorithms=None, **kwargs):
        if algorithms is None:
            header = jwt.get_unverified_header(token)
            algorithms = [header['alg']]

        return jwt.decode(token, key=key, algorithms=algorithms, **kwargs)


class BearerTokenExtractor:
    def extract(self) -> str:
        token = self._extract_token_from_header()
        if token is None:
            token = self._extract_token_from_query()
        return token

    @staticmethod
    def _extract_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.lower().startswith('bearer'):
            return auth.split(' ', 1)[1]

    @staticmethod
    def _extract_token_from_query():
        token = request.args.get('access_token')
        if token is not None:
            return token


class JwtManager:
    AUTHORITIES = 'authorities'
    USERNAME = 'username'
    USERINFO = 'userinfo'

    def __init__(self, public_key=None):
        self.public_key = public_key
        self.token_extractor = BearerTokenExtractor()

    def authenticate(self):
        access_token_value = self.token_extractor.extract()

        if access_token_value is not None:
            payload = self._decode(access_token_value)
            authorities = payload.get(self.AUTHORITIES)
            username = payload.get(self.USERNAME)
            userinfo = payload.get(self.USERINFO)
            user = User(username=username, authorities=authorities)
            if userinfo and isinstance(userinfo, dict):
                for k, v in userinfo.items():
                    setattr(user, k, v)

            SecurityContext.set_user(user)

    def _decode(self, access_token_value):
        try:
            payload = JwtHelper.decode_jwt(
                access_token_value,
                self.public_key,
            )
        except Exception as e:
            raise InvalidTokenError(e)
        return payload
